/* 
 *   Creation Date: <97/07/26 18:23:02 samuel>
 *   Time-stamp: <2003/05/22 13:24:27 samuel>
 *   
 *	<emulation.S>
 *	
 *	Low-level emulation of some privileged instructions
 *   
 *   Copyright (C) 1997-2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */
	
MACRO(INC_NIP, [
	addi	r6,r6,4
])

	
/************************************************************************/
/*	program exception vector					*/
/************************************************************************/
	
	//////////////////////////////////////////////////////////////
	// exception preamble
	//	r1		stack (mregs)
	//	r6,r7		nip / srr1
	//	cr4-7		flag_bits
	//
	// Saved r0-r5 (NOT SAVED: ctr, xer)

VECTOR_( 0x700, "Program", secint_bad, mac_entry_test )
	EXCEPTION_PREAMBLE			// 46 cycles on a G3
	TRACE(0x700, "Program")

	mtcrf	0xf0,r7				// put flag bits into cr0-cr3	
	bt	14,mac_program_trap
	GET_INSTR_OPCODE			// r6=nip, r2-r3, r4=opcode (cr not touched)
	bt+	13,emulate_priv_instr
	bt+	12,emulate_illegal_instr
	bt+	11,mac_program_trap		// bit 11 = fpu_exception

	// we should not come here
	MAC_EXIT_SAVE( RVEC_UNUSUAL_PROGRAM_EXCEP )

	
unhandled_priv_inst:
	MAC_EXIT_SAVE( RVEC_PRIV_INST )		// r4 = opcode

emulate_illegal_instr:
	// mfspr/mtspr 959 is illegal on 7455
	rlwinm	r0,r4,0,11,5			// clear reg field
	lis	r2,0x7c1f
	rlwinm	r0,r0,0,~0x100			// clear mtspr/mfspr bit
	ori	r2,r2,0xeaa6
	cmpw	r0,r2
	beq	emulate_priv_instr		// mfspr/mtspr 959
#ifdef EMULATE_603
	bl	emulate_603_instr
#endif
	MAC_EXIT_SAVE( RVEC_ILLEGAL_INST )	// r4 = opcode


/************************************************************************/
/*	mac exceptions							*/
/************************************************************************/
	
	/////////////////////////////////////////////////////////////
	// mac_trap
	//
	//	r2	exception vector
	//	r6	nip
	//	r7	reason bits (0-15)
	//
	// r0,r2-r5 may be modified

mac_interrupt_trap:
	li	r2,0x500
	li	r7,0			// no reason bits
	b	mac_trap

mac_dec_trap:
	li	r2,0x900
	crclr	FBIT_DecINT
	li	r7,0			// no reason bits
	b	mac_trap				

	balign_32
mac_program_trap:			// the reason bits already in r7 are used
	__BUMP("mac_trap")
	li	r2,0x700
mac_trap:
	lwz	r3,xMSR(r1)
	stw	r6,xSRR0(r1)		// mac-srr0 = nip
	mr	r4,r3			// copy xMSR to SRR1
	mr	r6,r2			// new nip
	rlwinm	r5,r3,25+1,31,31	// put MSR_IP (bit 25) in r5[31]
	rlwimi	r4,r7,0,0,15		// copy reason bits from r7 to SRR1
	rlwimi	r4,r3,0,6,6		// copy MSR_VEC bit of xMSR to SRR1
	stw	r4,xSRR1(r1)		// srr1 = (msr & (0xffff|MSR_VEC)) | (srr1 & ~(0xffff|MSR_VEC))
	rlwinm	r4,r3,0,25,25		// copy MSR_IP to xMSR
	neg	r5,r5			// r5 = 0xffffffff * MSR_IP
	rlwimi	r4,r3,0,19,19		// copy MSR_ME to xMSR
	rlwimi	r6,r5,0,0,11		// insert exception prefix
	stw	r4,xMSR(r1)
	bl	_msr_altered_exception

	GET_TICK_CNT(entry,"mac-expception")
	b	exception_return	// never singlestep exceptions

rvec_trace_trap:	
	MAC_EXIT_SAVE( RVEC_TRACE_TRAP )


/************************************************************************/
/*	decode privileged instruction					*/
/************************************************************************/

#define OPCODE(x,y)     ((x<<10)+y)
	
	/////////////////////////////////////////////////////////////
	// emulate_priv_instr
	//	r4	opcode
	//	r6,r7	nip/srr1
	//
	// r0,r2-r5,lr free

	balign_32
emulate_priv_instr:
	lwz	r3,xMSR(r1)
	rlwinm	r2,r4,32-1,22,31
	rlwinm	r5,r4,14,24,28			// r5 = reg_num << 3 (instr. bits 6-10)
	rlwimi	r2,r4,16,16,21			// r2 = opcode
	andi.	r3,r3,MSR_PR			// only emulate in supervisor mode
	bne-	mac_program_trap
	
	GET_TICK_CNT(entry,"get_inst")
	BUMP("decode_priv_inst")

	cmpwi	cr1,r2,OPCODE(31,339)		// mfspr
	beq-	cr1,emulate_mfspr

	cmpwi	cr2,r2,OPCODE(31,467)		// mtspr
	beq-	cr2,emulate_mtspr

	cmpwi	cr3,r2,OPCODE(31,83)		// mfmsr
	beq-	cr3,emulate_mfmsr

	cmpwi	cr0,r2,OPCODE(31,146)		// mtmsr
	beq-	cr0,emulate_mtmsr

	cmpwi	cr1,r2,OPCODE(19,50)		// rfi
	beq-	cr1,emulate_rfi

	cmpwi	cr2,r2,OPCODE(31,595)		// mfsr
	beq-	cr2,emulate_mfsr

	cmpwi	cr3,r2,OPCODE(31,659)		// mfsrin
	beq-	cr3,emulate_mfsrin

	cmpwi	cr0,r2,OPCODE(31,210)		// mtsr
	beq-	cr0,emulate_mtsr

	cmpwi	cr1,r2,OPCODE(31,242)		// mtsrin
	beq-	cr1,emulate_mtsrin

	cmpwi	cr2,r2,OPCODE(31,306)		// tlbie
	beq-	cr2,emulate_tlbie

	cmpwi	cr3,r2,OPCODE(31,566)		// tlbsync
	beq-	cr3,emulate_tlbsync

	cmpwi	cr0,r2,OPCODE(31,467)		// dcbi
	beq-	cr0,emulate_dcbi
#ifdef EMULATE_603
	cmpwi	cr1,r2,OPCODE(31,978)		// tlbld
	beq-	cr1,emulate_tlbld

	cmpwi	cr2,r2,OPCODE(31,1010)		// tlbli
	beq-	cr2,emulate_tlbli
#endif
	// Program-trap, illegal instruction
	b	unhandled_priv_inst		// r4 = opcode


#ifdef EMULATE_603
emulate_603_instr:
	rlwinm	r2,r4,32-1,22,31
	rlwimi	r2,r4,16,16,21			// r2 = opcode

	cmpwi	cr0,r2,OPCODE(31,978)		// tlbld
	beq	cr0,2f
	cmpwi	cr1,r2,OPCODE(31,1010)		// tlbli
	beq	cr1,2f
	cmpwi	cr2,r2,OPCODE(31,339)		// mfspr
	beq	cr2,1f
	cmpwi	cr3,r2,OPCODE(31,467)		// mtspr
	bnelr	cr3
1:	rlwinm	r3,r4,32-16,27,31
	rlwimi	r3,r4,32-6,22,26		// r3 = spr#
	cmpwi	r3,976				// first 603 SPR
	bltlr
	cmpwi	r3,982				// last 603 SPR
	bgtlr
2:
	rlwinm	r7,r7,0,14,10			// clear bit 11-13
	oris	r7,r7,0x4			// set bit 13
	b	emulate_priv_instr
	
#endif /* EMULATE_603 */

			
/************************************************************************/
/*	mac register access						*/
/************************************************************************/

MACRO(LOAD_EMUGPR_IM, [dreg,ereg], [
.if _ereg <= 7
	lwz	_dreg,xGPR0+_ereg[]*4(r1)
.else
	mr	_dreg,rPREFIX[]_ereg
.endif
])

MACRO(STORE_EMUGPR_IM, [sreg,ereg], [
.if _ereg <= 7
	stw	_sreg,xGPR0+_ereg[]*4(r1)
.else
	mr	rPREFIX[]_ereg, _sreg
.endif
])
	
	balign_32
store_gpr_table:
mFORLOOP([i],0,31,[
	STORE_EMUGPR_IM r0,i
	blr
])

load_gpr_table:
mFORLOOP([i],0,31,[
	LOAD_EMUGPR_IM r0,i
	blr
])

MACRO(EMU_LOAD_GPR, [reg, scr], [
	LI_PHYS( _scr, load_gpr_table )
	add	rPREFIX[]_scr,_reg,rPREFIX[]_scr
	mtlr	rPREFIX[]_scr
	blrl
])

MACRO(EMU_STORE_GPR, [reg, scr], [
	LI_PHYS( _scr, store_gpr_table )
	add	rPREFIX[]_scr,_reg,rPREFIX[]_scr
	mtlr	rPREFIX[]_scr
	blrl
])
	

/************************************************************************/
/*	HOTSPOT search (DEBUG)						*/
/************************************************************************/

//#define HOTSPOT_SEARCH

#ifdef HOTSPOT_SEARCH
hotspot_rec:
	LI_PHYS( R3,r__hotspot_table )
	LI_PHYS( R4,r__hotspot_table_end )
	addi	r3,r3,-8
1:	
	lwzu	r2,8(r3)
	cmpwi	r2,0
	bne	2f

	// New entry
	stw	r6,0(r3)		// NIP
	li	r2,0
	stw	r2,4(r3)		// count
	blr

2:	cmpw	r6,r2			// NIP match?
	bne	3f
	lwz	r2,4(r3)		// inc count one
	addi	r2,r2,1
	stw	r2,4(r3)
	blr
3:	
	cmpw	r3,r4
	blt	1b
	blr

GLOBAL_SYMBOL(r__hotspot_table):
	.space	4096,0
GLOBAL_SYMBOL(r__hotspot_table_end): 
	
#endif

	
/************************************************************************/
/*	instruction emulation						*/
/************************************************************************/

	//////////////////////////////////////////////////////////
	// emulate_xxxxx
	//	r4	opcode
	//	r5	regnum<<3 (from opcode bits 6-10)
	//
	// May modify: r0,r2-r5 (lr)

/********************************************************************/
emulate_mfmsr:
	lwz	r0,xMSR(r1)
	EMU_STORE_GPR r5, /**/ R2
	GET_TICK_CNT( entry, "mfmsr" )
	b	emulation_done

/********************************************************************/
emulate_mfspr:
	BUMP("emulate_mfspr")
	rlwinm	r2,r4,32-14,25,29
	rlwimi	r2,r4,32-4,20,24		// r2 = spr# << 2
	addi	r3,r1,xSPR_BASE
	lwzx	r0,r2,r3			// value in r0
	addi	r3,r1,K_SPR_HOOKS
	lwzx	r3,r2,r3			// hook in r3
	cmpwi	r3,0
	beq-	1f				// non-zero?
	mtlr	r3
	blrl
1:
	EMU_STORE_GPR r5, /**/ R3
	GET_TICK_CNT(entry,"mfspr")
	b	emulation_done

/********************************************************************/
emulate_mtspr:
	BUMP("emulate_mtspr")
	rlwinm	r2,r4,32-14,25,29
	rlwimi	r2,r4,32-4,20,24		// r2 = spr# << 2
	EMU_LOAD_GPR r5, /**/ R3		// value in r0
	addi	r3,r1,K_SPR_HOOKS
	lwzx	r3,r2,r3			// hook in r3
	cmpwi	r3,0
	beq-	1f				// non-zero?
	addi	r3,r3,4				// branch to hook +4
	mtlr	r3
	blrl
1:	
	addi	r3,r1,xSPR_BASE
	stwx	r0,r2,r3			// value in r0
	GET_TICK_CNT(entry,"mtspr")
	b	emulation_done


/********************************************************************/
	balign_32
emulate_mtmsr:
#ifdef HOTSPOT_SEARCH
	bl	hotspot_rec
#endif
	cror	4+eq,FBIT_MsrModified, FBIT_IRQMightBePending
	lwz	r3,xMSR(r1)
	LI_PHYS( r2, load_gpr_table )
	add	r2,r5,r2
	mtlr	r2
	blrl
	xor	r2,r3,r0			// r2 = altered bits
	cror	4+eq, 4+eq, FBIT_DecINT
	stw	r0,xMSR(r1)
	rlwinm.	r2,r2,0,17,15			// Clear MSR_EE toggle (bit16)
	
	INC_NIP
	crorc	eq,4+eq,eq			// FBIT_xxx || !eq
	beql-	_msr_altered

	GET_TICK_CNT(entry, "mtmsr")
	BUMP("emulate_mtmsr")
		
	// doze?
	lwz	r0,xMSR(r1)
	rlwinm.	r0,r0,0,13,13			// MSR_POW (bit13) set?
	beq+	emulation_done_noinc
	MAC_EXIT_SAVE( RVEC_MSR_POW )		// POW 1 => doze


/********************************************************************/
emulate_rfi:
	BUMP("rfi")
	lwz	r6,xSRR0(r1)			// new nip = SRR0
	lwz	r3,xMSR(r1)
	lwz	r4,xSRR1(r1)
	rlwimi	r3,r4,0,16,31			// msr = (msr & ~0xffff) | (SRR1 & 0xffff)
	rlwimi	r3,r4,0,6,6			// MSR_VEC must be copied too
#ifdef EMULATE_603
	rlwimi	r3,r4,0,14,14			// MSR_TGPR (603 specific)
#endif
	stw	r3,xMSR(r1)
	bl	_msr_altered			// r0,r2-r5, r7 [srr1] updated

	GET_TICK_CNT(entry,"rfi")
	lwz	r3,K_BREAK_FLAGS(r1)		// break at rfi support
	andi.	r3,r3,BREAK_RFI
	beq+	exception_return
	li	r4,BREAK_RFI			// r4 = flag causing the break
	MAC_EXIT_SAVE( RVEC_BREAK )


/********************************************************************/
#ifdef EMULATE_603
emulate_tlbli:
	LI_VIRT( r3, EXTERN(do_tlbli) )
	b	1f
emulate_tlbld:	
	LI_VIRT( r3, EXTERN(do_tlbld) )
	b	1f
emulate_tlbie:
	LI_VIRT(r3,EXTERN(do_tlbie) )
1:	INC_NIP
	rlwinm	r5,r4,32-8,24,28		// r5 = #B << 3
	EMU_LOAD_GPR r5, /**/ R2		// value ret. in r0
	mr	r4,r0				// r4 = ea
	b	call_kernel_save
#else
emulate_tlbie:
	b	emulation_done
#endif /* EMULATE_603 */
	
/********************************************************************/
emulate_tlbsync:
	b	emulation_done

/********************************************************************/
emulate_dcbi:
	b	unhandled_priv_inst		// r4 = opcode

/********************************************************************/
emulate_mfsrin:
	rlwinm	r2,r4,32-8,24,28		// r2 = #B << 3
	EMU_LOAD_GPR r2, /**/ R3		// r0 = reg B
	rlwinm	r3,r0,6,26,29			// r3 = #sr << 2
	b	1f
emulate_mfsr:
	rlwinm	r3,r4,32-14,26,29		// r3 = #sr << 2
1:	addi	r2,r1,xSEGR_BASE
	lwzx	r0,r3,r2
	EMU_STORE_GPR r5, /**/ R2
	GET_TICK_CNT(entry, "mfsr")
	BUMP("mfsr_")
	b	emulation_done


/************************************************************************/
/*	SPR - emulation							*/
/************************************************************************/

	////////////////////////////////////////////////////////////
	// read (offset 0)
	//	r0	spr_value
	//	r2	spr# << 2	*not used*
	//	r5	dreg << 3
	//
	// write (offset 4)
	//	r0	value_in_gpr
	//	r2	spr# << 2	*used on return*
	//
	// Safe to modify:	r3,r4, (write r5), lr 
	// NOT SAVED:		ctr, xer

	
/********************************************************************/
spr_read_only:	
	blr				// allow read
	b	emulation_done		// ignore write	
	
/********************************************************************/
unhandled_spr:
	b	unhandled_spr_read	// read hook	(offs 0)
unhandled_spr_write:			// write hook	(offs 4)
	srwi	r4,r2,2
	mr	r5,r0
	// r4 = spr#
	// r5 = register-value
	MAC_EXIT_SAVE( RVEC_SPR_WRITE )
unhandled_spr_read:
	srwi	r4,r2,2
	srwi	r5,r5,3
	// r4 = spr#
	// r5 = dest gpr
	MAC_EXIT_SAVE( RVEC_SPR_READ )

/********************************************************************/
spr_bat:
	blr				// read has no side-effects
	INC_NIP
	LI_VIRT( r3, EXTERN(do_mtbat) )
	bl	save_middle_regs	// Must do this before touching r6-r12
	srwi	r4,r2,2			// r4 = spr#
	mr	r5,r0			// r5 = value
	li	r6,0			// not forced
	b	call_kernel

/********************************************************************/
spr_sdr1:
	blr				// read has no side-effects
	INC_NIP
	LI_VIRT( r3, EXTERN(do_mtsdr1) )
	mr	r4,r0			// r4 = value
	b	call_kernel_save


/************************************************************************/
/*	handle MSR changes						*/
/************************************************************************/

	balign_32
_msr_altered_exception:	
	lwz	r7,K_MSR(r1)				// r7 = K_MSR
	li	r5,MSR_FE0 | MSR_FE1 | MSR_SE | MSR_FP | MSR_BE
	rlwinm	r7,r7,0,7,5				// Clear MSR_VEC (bit 6)

	crclr	FBIT_MsrModified
		
	lwz	r4,K_CUR_SR_BASE(r1)			// detect sr change
	addi	r3,r1,K_UNMAPPED_SR_BASE		// set unmapped context
	stw	r3,K_CUR_SR_BASE(r1)
	stw	r3,K_SR_DATA(r1)
	stw	r3,K_SR_INST(r1)
	sub.	r3,r3,r4				// has cur_sr_base changed?
	crorc	FBIT_LoadSegreg,FBIT_LoadSegreg,eq	// if so, reload segment registers

	andc	r7,r7,r5				// Clear most MSR bits
	
	lwz	r3,K_TRANSL_DBAT0U(r1)
	lwz	r4,K_TRANSL_DBAT0L(r1)
	stw	r3,K_DBAT0U(r1)
	stw	r4,K_DBAT0L(r1)

	stw	r7,K_MSR(r1)

	bflr+	FBIT_DbgSingleStep			// predict taken
	ori	r7,r7,MSR_SE
	stw	r7,K_MSR(r1)
	blr


	////////////////////////////////////////////////////////////
	// _msr_altered
	//
	//	r6, r7:		nip / srr1
	//
	// modifies: r0,r2-r5 (r7 updated)
	//
	// This code should be called just before exception return.
	// It might throw a mac DEC or interrupt exception.

	balign_32
msr_test_for_irq_trap:
	rlwinm.	r3,r0,0,16,16				// MSR_EE
	beq	1f
	lwz	r3,xIRQ(r1)
	cmpwi	r3,0
	bne	mac_interrupt_trap
	crclr	FBIT_IRQMightBePending			// clear flag since xIRQ is 0
	bt-	FBIT_DecINT,mac_dec_trap
	b	1f

	balign_16
msr_test_for_dec_trap:
	BUMP("test_for_dec_trap")
	rlwinm. r3,r0,0,16,16				// MSR_EE
	beq	1f
	BUMP("dec_trap")
	b	mac_dec_trap

	balign_32
_msr_altered:
	lwz	r0,xMSR(r1)				// [r0] = xMSR, IR,DR in bits 26,27
	lwz	r2,K_CUR_SR_BASE(r1)			// save current sr base in r2

	// throw an exception?
	bt-	FBIT_IRQMightBePending,msr_test_for_irq_trap	// r0 = xMSR, modifies r3
	bt-	FBIT_DecINT,msr_test_for_dec_trap		// r0 = xMSR, modifies r3
1:
	crclr	FBIT_MsrModified

	// set sr bases from K_MSR_SR_TABLE[ PR IR DR ]
	rlwinm	r5,r0,0,26,27				// IR,DR bits
	rlwimi	r5,r0,32-8,25,25			// PR bit
	addi	r5,r5,K_MSR_SR_TABLE
	lwzux	r3,r5,r1
	lwz	r4,4(r5)
	lwz	r5,8(r5)
	stw	r3,K_CUR_SR_BASE(r1)			// new sr base in r3 (used below)
	stw	r4,K_SR_DATA(r1)
	stw	r5,K_SR_INST(r1)

	sub.	r3,r2,r3				// has cur_sr_base changed?
	crorc	FBIT_LoadSegreg,FBIT_LoadSegreg,eq	// if so, reload segment registers

	li	r2,MSR_FP | MSR_FE0 | MSR_FE1		// r2 = MSR_FEx bits
	
	// in splitmode?
	rlwinm	r3,r0,0,26,26				// MSR_IR
	rlwinm	r4,r0,1,26,26				// MSR_DR (bit 27)
	sub.	r4,r4,r3				// MSR_IR == MSR_DR?
	crnot	FBIT_InSplitmode,eq			// other split bits set below

	lwz	r3,K_TRANSL_DBAT0U(r1)
	lwz	r4,K_TRANSL_DBAT0L(r1)
	andc	r7,r7,r2				// r7 = (K_MSR & ~MSR_FEx)
	mfcr	r5					// r5 = CR (we want FBIT_FPUInUse)
	stw	r3,K_DBAT0U(r1)				// set by prepare_splitmode but
	stw	r4,K_DBAT0L(r1)				// we don't want an extra branch
	rlwinm	r3,r5,FBIT_FPUInUse+1,31,31		// r3 = 1 * FPUInUse

	crmove	FBIT_PrepareSplitmode,FBIT_InSplitmode

	// copy SE,BE and FP,FEx bits to physical msr, clear MSR_VEC
	neg	r3,r3					// r3 = 0xffffffff * FPUInUse
	rlwimi	r7,r0,0,21,22				// insert MSR_SE and MSR_BE
	and	r3,r3,r2				// r3 = (MSR_FEx mask) * FPUInUse
	rlwinm	r7,r7,0,7,5				// Clear MSR_VEC (bit 6)
	and	r3,r0,r3				// r3 = (xMSR & MSR_FEx) * FPUInUse
	or	r7,r7,r3				// set FPx bits in K_MSR
		
	cror	FBIT_LoadSegreg,FBIT_LoadSegreg,FBIT_InSplitmode

#ifdef EMULATE_603
	// 603 alternate GPR emulation
	bf+	FBIT_603_AltGPR,7f
	rlwinm.	r5,r0,0,14,14				// MSR_TGPR (temp GPR remap)
	bne+	7f
	lwz	r2,xGPRSAVE0_603(r1)
	lwz	r3,xGPRSAVE1_603(r1)
	lwz	r4,xGPRSAVE2_603(r1)
	lwz	r5,xGPRSAVE3_603(r1)
	stw	r2,xGPR0(r1)
	stw	r3,xGPR1(r1)
	stw	r4,xGPR2(r1)
	stw	r5,xGPR3(r1)
	crclr	FBIT_603_AltGPR
7:
#endif
	// singlestepping stuff
	rlwinm.	r5,r0,0,21,21				// MSR_SE set (bit 21)?
	crorc	FBIT_SingleStep,FBIT_DbgSingleStep,eq	// update FBIT_SingleStep
	stw	r7,K_MSR(r1)
	bflr+	FBIT_DbgSingleStep			// predict taken
	ori	r7,r7,MSR_SE
	stw	r7,K_MSR(r1)
	blr
	
	// msr_altered( kv )
GLOBAL_SYMBOL(r__msr_altered):
	lwz	r5,xFLAG_BITS(r3)
	ori	r5,r5,fb_MsrModified
	stw	r5,xFLAG_BITS(r3)
	blr

	
/************************************************************************/
/*	initialize special purpose register table			*/
/************************************************************************/

MACRO(SPR_HOOK,	[spr, hook],	[ LI_PHYS( R8, _hook ) ; stw r8,(((_spr)*4)+K_SPR_HOOKS)(r1) ])
MACRO(ZERO_SPR_HOOK, [spr],	[ stw r7,(((_spr)*4)+K_SPR_HOOKS)(r1) ])

initialize_spr_table:
	LI_PHYS( R7, unhandled_spr )
	addi	r8,r1,K_SPR_HOOKS-4
	li	r9,1024
	mtctr	r9
1:	stwu	r7,4(r8)
	bdnz	1b

	li	r7,0
#ifdef EMULATE_603
	mFOREACH(x, [IMISS, ICMP, DMISS, DCMP, HASH1, HASH2, RPA],
		[ZERO_SPR_HOOK x ; ])
#endif
	mFOREACH(x, [SRR0, SRR1, SPRG0, SPRG1, SPRG2, SPRG3, DSISR, DAR, EAR, HID0],
			[ZERO_SPR_HOOK x ; ])

	// for now...
	ZERO_SPR_HOOK	TBWU
	ZERO_SPR_HOOK	TBWL

	// SPRs that have side effects
	SPR_HOOK	PVR,	spr_read_only
	SPR_HOOK	SDR1,	spr_sdr1
	SPR_HOOK	DEC,	spr_dec

	// BATs
	mFORLOOP([nn],0,15,[
	 SPR_HOOK	nn+IBAT0U, spr_bat
	])
	blr

	
/************************************************************************/
/*	initialize msr segment register table				*/
/************************************************************************/
	
	/////////////////////////////////////////////////////////////
	// initialize_sr_offs_table
	//
	// Copy sr_offs_table to K_MSR_SR_TABLE
	// r1 is added to each element
	
initialize_msr_sr_table:
	mflr	r8			// Get address of table
	bl	sr_offs_table
	mflr	r3
	mtlr	r8

	li	r5,4*8			// #words in table	
	mtctr	r5
	addi	r3,r3,-4
	addi	r4,r1,K_MSR_SR_TABLE-4
1:
	lwzu	r6,4(r3)
	add	r6,r6,r1		// And add r1
	stwu	r6,4(r4)
	bdnz	1b
	blr

	// Used to construct msr_sr_table (mbase is added)
sr_offs_table:
	blrl
	/*	K_CUR_SR_BASE,		K_SR_DATA_BASE,		K_SR_INST_BASE,		dummy */
	
	.long	K_UNMAPPED_SR_BASE,	K_UNMAPPED_SR_BASE,	K_UNMAPPED_SR_BASE,	0
	.long	K_SPLIT_SR_BASE,	K_SV_SR_BASE,		K_UNMAPPED_SR_BASE,	0	/* DR */
	.long	K_SPLIT_SR_BASE,	K_UNMAPPED_SR_BASE,	K_SV_SR_BASE,		0	/* IR */
	.long	K_SV_SR_BASE,		K_SV_SR_BASE,		K_SV_SR_BASE,		0	/* DR|IR */

	.long	K_UNMAPPED_SR_BASE,	K_UNMAPPED_SR_BASE,	K_UNMAPPED_SR_BASE,	0	/* PR */
	.long	K_SPLIT_SR_BASE,	K_USER_SR_BASE,		K_UNMAPPED_SR_BASE,	0	/* PR|DR */
	.long	K_SPLIT_SR_BASE,	K_UNMAPPED_SR_BASE,	K_USER_SR_BASE,		0	/* PR|IR */
	.long	K_USER_SR_BASE,		K_USER_SR_BASE,		K_USER_SR_BASE,		0	/* PR|DR|IR */
