/*   -*- asm -*-
 *
 *   Creation Date: <2001/01/28 20:33:22 samuel>
 *   Time-stamp: <2003/03/07 16:35:36 samuel>
 *   
 *	<molasm.h>
 *	
 *	Utility assembly macros
 *   
 *   Copyright (C) 2001, 2002, 2003 Samuel Rydh (samuel@ibrium.se)
 *   
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License
 *   as published by the Free Software Foundation
 *   
 */

#ifndef _H_MOLASM
#define _H_MOLASM

/************************************************************************/
/*	SPRG usage							*/
/************************************************************************/

/* Darwin and Linux uses the sprg's differently. Linux uses sprg0/1 in
 * the exception vectors while Darwin uses sprg2/3.
 */
#ifdef __linux__
define([mfsprg_a0], [mfsprg0])
define([mfsprg_a1], [mfsprg1])
define([mfsprg_a2], [mfsprg2])
define([mfsprg_a3], [mfsprg3])
define([mtsprg_a0], [mtsprg0])
define([mtsprg_a1], [mtsprg1])
define([mtsprg_a2], [mtsprg2])
define([mtsprg_a3], [mtsprg3])
#else
define([mfsprg_a0], [mfsprg2])
define([mfsprg_a1], [mfsprg3])
define([mfsprg_a2], [mfsprg0])
define([mfsprg_a3], [mfsprg1])
define([mtsprg_a0], [mtsprg2])
define([mtsprg_a1], [mtsprg3])
define([mtsprg_a2], [mtsprg0])
define([mtsprg_a3], [mtsprg1])
#endif
	
/************************************************************************/
/*	Vector entry point definitions					*/
/************************************************************************/

#define MOL_SPRG2_MAGIC		0x1779

/*
 * This code uses the dynamic linkage/action symbol functionality of 
 * the MOL kernel loader to automatically install the hooks. Refer to 
 * hook.c for the actual implementation.
 */

/* Description of Action_RELOC_HOOK:
 *
 *	.long	Action_RELOC_HOOK
 *	.long	vector
 *	.long	#bytes to copy to lowmem
 *	.long	offset to vret function
 *	.long	offset to vector entry
 *	-- offsets are calculated from here --
 */

#define VECTOR_HOOK( v )					\
	.long	EXTERN( Action_RELOC_HOOK )			;\
	.long	v						;\
	.long	vhook_end_##v - vhook_##v			;\
	.long	vret_##v - vhook_##v				;\
	.long	vhook_entry_##v - vhook_##v			;\
vhook_##v:							;\
	mtsprg_a0 r3						;\
	addis	r3,0,0	   /* [1] hook address inserted */	;\
	mtsprg_a1 r1						;\
	ori	r3,r3,0	   /* [3] at module initialization */	;\
	mflr	r1						;\
	mtlr	r3						;\
	blr							;\
								;\
vret_##v:							;\
	nop	/* overwritten instruction is inserted here */	;\
	ba	v + 0x4						;\
vhook_end_##v:							;\
	balign_32						;\
vhook_entry_##v:						;
		
/* SPRG0,1 = saved r3,r1, r1 = saved lr */
#define VECTOR_(v, dummy_str, secondary, not_mol_label )	\
	/* not mol return */					;\
rLABEL(lab_716):						;\
	mtcr	r3						;\
	mfsprg_a1 r1						;\
	mfsprg_a0 r3						;\
rLABEL(lab_717): ; continue_trap_##v:				;\
	.long	EXTERN( Action_VRET ) + v  /* ba vret_xxx */	;\
								;\
	/* secondary interrupt */				;\
rLABEL(lab_718):						;\
	li	r3,v						;\
	b	secondary					;\
								;\
	VECTOR_HOOK( v )					;\
								;\
	/* entrypoint */					;\
	mtlr	r1						;\
	mfcr	r3						;\
	mfsprg_a2 r1						;\
	cmpwi	r1,MOL_SPRG2_MAGIC				;\
	bne-	not_mol_label					;\
soft_603_entry_##v:						;\
	mfsrr1	r1						;\
	rlwinm.	r1,r1,0,17,17	/* MSR_PR set? */		;\
	mfsprg_a3 r1						;\
	beq-	lab_718b	/* if not, take a secondary trap? */	;
	
#define VECTOR(v, dummy_str, secondary) VECTOR_(v, dummy_str, secondary, lab_716b )

#define TAKE_EXCEPTION_( v ) 					\
	bl	take_exception					;\
	.long	EXTERN( Action_VRET ) + v			;

#define TAKE_EXCEPTION						\
	bl	take_exception					;\
	b	lab_717b


/************************************************************************/
/*	603 vector HOOKs (r0-r3, cr0 saved by hardware)			*/
/************************************************************************/

#define VECTOR_HOOK_603( v )					\
	.long	EXTERN( Action_RELOC_HOOK )			;\
	.long	v						;\
	.long	vhook_end_##v - vhook_##v			;\
	.long	vret_##v - vhook_##v				;\
	.long	vhook_entry_##v - vhook_##v			;\
vhook_##v:							;\
	mfsprg_a2 r1						;\
	addis	r3,0,0	   /* [1] hook address inserted */	;\
	cmpwi	r1,MOL_SPRG2_MAGIC				;\
	ori	r3,r3,0	   /* [3] at module initialization */	;\
	bne	vret_##v					;\
	mfctr	r0						;\
	mtctr	r3						;\
	bctr							;\
								;\
vret_##v:							;\
	nop	/* overwritten instruction is inserted here */	;\
	ba	v + 0x4						;\
vhook_end_##v:							;\
	balign_32						;\
vhook_entry_##v:						;

/* r0 = saved ctr */
#define VECTOR_603(v, dummy_str )				\
	VECTOR_HOOK_603( v )					;\
	/* entrypoint goes here */				;

/* all register are assumed unmodified here */
#define SOFT_VECTOR_ENTRY_603( v )				\
	mtsprg_a0 r3						;\
	mtsprg_a1 r1						;\
	mfcr	r3						;\
	b	soft_603_entry_##v				;


/************************************************************************/
/*	FUNCTION_HOOK							*/
/************************************************************************/

#define FHOOK( symind )						\
	.long	EXTERN( Action_HOOK_FUNCTION )			;\
	.long	symind						;\
	.long	fhook_end_##symind - fhook_##symind		;\
	.long	fret_##symind - fhook_##symind			;\
fhook_##symind:							;\
	mflr	r10						;\
	addis	r9,0,0		/* [1] address inserted */	;\
	ori	r9,r9,0		/* [2] at runtime */		;\
	mtctr	r9						;\
	bctrl							;\
	mtlr	r10						;\
fret_##symind:							;\
	nop	/* overwritten instruction is inserted here */	;\
	nop	/* return (through a relative branch) */	;\
fhook_end_##symind:						;


/************************************************************************/
/*	Utility								*/
/************************************************************************/

MACRO(LOAD_VARIABLE, [reg, offs], [
	lis	_reg,HA(k_mol_stack + _offs)
	lwz	_reg,LO(k_mol_stack + _offs)(_reg)
])

MACRO(SET_SESSION_TABLE, [reg], [
	lis	_reg,HA(EXTERN(Symbol_SESSION_TABLE))
	addi	_reg,_reg,LO(EXTERN(Symbol_SESSION_TABLE))
])


/************************************************************************/
/*	GPR save / restore						*/
/************************************************************************/

MACRO(xGPR_SAVE, [reg], [
	stw	rPREFIX[]_reg,(xGPR0 + _reg*4)(r1)
])
		
MACRO(xGPR_LOAD, [reg], [
	lwz	rPREFIX[]_reg,(xGPR0 + _reg*4)(r1)
])

/************************************************************************/
/*	FPU misc							*/
/************************************************************************/

MACRO(ENABLE_MSR_FP, [scr], [
	mfmsr	_scr
	ori	_scr,_scr,MSR_FP
	mtmsr	_scr
	isync
])
		
/************************************************************************/
/*	Segment registers						*/
/************************************************************************/

MACRO(LOAD_SEGMENT_REGS, [base, scr, scr2], [
	mFORLOOP([i],0,7,[
		lwz	_scr,eval(i * 8)(_base)
		lwz	_scr2,eval((i * 8)+4)(_base)
		mtsr	srPREFIX[]eval(i*2),_scr
		mtsr	srPREFIX[]eval(i*2+1),_scr2
	])
])

MACRO(SAVE_SEGMENT_REGS, [base, scr, scr2], [
	mFORLOOP([i],0,7,[
		mfsr	_scr,srPREFIX[]eval(i*2)
		mfsr	_scr2,srPREFIX[]eval(i*2+1)
		stw	_scr,eval(i * 8)(_base)
		stw	_scr2,eval((i * 8) + 4)(_base)
	])
])

/************************************************************************/
/*	BAT register							*/
/************************************************************************/

MACRO(SAVE_DBATS, [varoffs, scr1], [
	mfpvr	_scr1
	srwi	_scr1,_scr1,16
	cmpwi	r3,1
	beq	9f
	mFORLOOP([nn],0,7,[
		mfspr	_scr1, S_DBAT0U + nn
		stw	_scr1,(_varoffs + (4 * nn))(r1)
	])
9:
])
	
MACRO(SAVE_IBATS, [varoffs, scr1], [
	mFORLOOP([nn],0,7,[
		mfspr	_scr1, S_IBAT0U + nn
		stw	_scr1,(_varoffs + (4 * nn))(r1)
	])
])

	
/************************************************************************/
/*	Physical/virtual conversion					*/
/************************************************************************/

	/* replaced with lis dreg,addr@ha ; addi dreg,dreg,addr@l */
#define LI_PHYS( dreg, addr ) \
	.long	EXTERN( Action_LI_PHYS ) + dreg		;\
	.long	addr - r__reloctable_start

#define LI_VIRT( dreg, addr ) \
	lis	dreg,HA(addr)			;\
	addi	dreg,dreg,LO(addr)

#define LI_PHYS_VARBASE( base ) \
	LI_PHYS( base, k_variable_base )

#define LI_VIRT_VARBASE( base ) \
	LI_VIRT( base, k_variable_base )

	/* syntax: lwz dreg, phys_addr(reg)
	 * replaced with addis dreg,reg,addr@ha ; lwz dreg,addr@lo(dreg).
	 */
#define LWZ_PHYSADDR_R( dreg, addr, reg )	;\
	.long	EXTERN( Action_LWZ_PHYSADDR_R ) + (dreg*32 + reg)	;\
	.long	addr - r__reloctable_start

#define LWZ_PHYS( dreg, addr )	;\
	LWZ_PHYSADDR_R( dreg, addr, 0 );

	/* syntax: tophys rD,rS */
MACRO(tophys, [dreg, sreg], [
	.long	EXTERN( Action_TOPHYS ) + _dreg * 32 + _sreg
])
	/* syntax: tovirt rD,rS */
MACRO(tovirt, [dreg, sreg], [
	.long	EXTERN( Action_TOVIRT ) + _dreg * 32 + _sreg
])
	
	
/************************************************************************/
/*	Performance Statistics						*/
/************************************************************************/

#ifdef PERFORMANCE_INFO
	define([_bump_ind_], 0)
		
#define __BUMP( str )					\
	.text	92					;\
debug_str_[]_bump_ind_:					;\
	.if (_bump_ind_ >= NUM_ASM_BUMP_CNTRS)		;\
	.print "** too many BUMP counters **" ; .fail 1	;\
	.endif						;\
	.ascii	str "\0"				;\
	balign_4					;\
	.text	90					;\
	.long	debug_str_[]_bump_ind_			;\
	.text						;\
	stw	r3,xDEBUG_SCR1(r1)			;\
	lwz	r3,(K_ASM_BUMP_CNTRS+4*_bump_ind_)(r1)	;\
	addi	r3,r3,1					;\
	stw	r3,(K_ASM_BUMP_CNTRS+4*_bump_ind_)(r1)	;\
	lwz	r3,xDEBUG_SCR1(r1)			;\
	define([_bump_ind_],eval(_bump_ind_+1))


	define([_tick_ind_], 0)
	
#define __ZERO_TICK_CNT(cntr)					\
	ifdef([##cntr##_ind_],[],[				\
	  define([##cntr##_ind_], _tick_ind_)			\
	  define([_tick_ind_], eval(_tick_ind_+1))		\
	])							\
	.if (_tick_ind_ > NUM_ASM_TICK_CNTRS)			;\
	.print "** too many TICK counters **" ; .fail 1		;\
	.endif							;\
	stw	r3,xDEBUG_SCR1(r1)				;\
	mftb	r3						;\
	stw	r3,(K_ASM_TICK_STAMPS + 4*cntr##_ind_)(r1)	;\
	lwz	r3,xDEBUG_SCR1(r1)
	
#define __GET_TICK_CNT(cntr, name) \
	.text	92						;\
debug_str_[]_bump_ind_:						;\
	.if (_bump_ind_ >= NUM_ASM_BUMP_CNTRS)			;\
	.print "** too many BUMP counters **" ; .fail 1		;\
	.endif							;\
	.ascii	name "_ticks\0"					;\
	balign_4						;\
	.text	90						;\
	.long	debug_str_[]_bump_ind_				;\
	.text							;\
	ifdef([##cntr##_ind_],[],[				\
	  define([##cntr##_ind_], _tick_ind_)			\
	  define([_tick_ind_], eval(_tick_ind_+1))		\
	])							\
	.if (_tick_ind_ > NUM_ASM_TICK_CNTRS)			;\
	.print "** too many TICK counters **" ; .fail 1		;\
	.endif							;\
	stw	r3,xDEBUG_SCR1(r1)				;\
	mftb	r3						;\
	stw	r4,xDEBUG_SCR2(r1)				;\
	lwz	r4,(K_ASM_TICK_STAMPS + 4*cntr##_ind_)(r1)	;\
	sub	r3,r3,r4					;\
	lwz	r4,(K_ASM_BUMP_CNTRS+4*_bump_ind_)(r1)		;\
	add	r4,r4,r3					;\
	stw	r4,(K_ASM_BUMP_CNTRS+4*_bump_ind_)(r1)		;\
	lwz	r3,xDEBUG_SCR1(r1)				;\
	lwz	r4,xDEBUG_SCR2(r1)				;\
	define([_bump_ind_],eval(_bump_ind_+1))

#endif /* PERFORMANCE_INFO */

#ifndef PERFORMANCE_INFO_LIGHT
#define BUMP(s)			__BUMP(s)
#define ZERO_TICK_CNT(c)	__ZERO_TICK_CNT(c)
#define GET_TICK_CNT(c, name)	__GET_TICK_CNT(c,name)
#else
#define BUMP(s)
#define ZERO_TICK_CNT(c)
#define GET_TICK_CNT(c, name)
#endif

#ifndef __BUMP
#define __BUMP(str)
#define __ZERO_TICK_CNT(cntr)
#define __GET_TICK_CNT(cntr, name)
#endif

#ifdef PERF_MONITOR
MACRO(PERF_MONITOR_GET, [
	stw	r5,xDEBUG_SCR1(r1)
	mfspr	r5,S_PMC2
	stw	r4,xDEBUG_SCR2(r1)
	mfmsr	r4
	ori	r4,r4,MSR_PE
	mtmsr	r4
	stw	r5,xDEBUG0(r1)
	li	r5,0
	mtspr	S_PMC2,r5
	lwz	r4,xDEBUG_SCR2(r1)
	lwz	r5,xDEBUG_SCR1(r1)
])
MACRO(PERF_MONITOR_SETUP, [scr], [
	LOADI	_scr,BIT(2) | BIT(3) | BIT(31)	// count in SV-mode if PM is zero.
	mtspr	S_MMCR0,_scr
	li	_scr,0
	mtspr	S_MMCR1,_scr
	li	_scr,0
	mtspr	S_PMC2,_scr
])
#else /* PERF_MONITOR */
#define PERF_MONITOR_GET
MACRO(PERF_MONITOR_SETUP, [scr], [])
#endif

	
/************************************************************************/
/*	D E B U G							*/
/************************************************************************/

MACRO(STOP_EMULATION, [val], [
	stw	r3,xDEBUG_SCR1(r1)
	li	r3,_val
	stw	r3,xKERNEL_DBG_STOP(r1)
	li	r3,1
	stw	r3,xINTERRUPT(r1)
	lwz	r3,xDEBUG_SCR1(r1)	
])

MACRO(DEBUG_TRACE, [num, dummy], [
	stw	r3,xDEBUG_SCR1(r1)
	lwz	r3,xDEBUG_TRACE(r1)
	addi	r3,r3,1
	stw	r3,xDEBUG_TRACE(r1)
	stw	r3,(xDEBUG0+4*_num)(r1)
	lwz	r3,xDEBUG_SCR1(r1)
])

MACRO(TRACE_VAL, [val, dummy], [
#if DBG_TRACE
	stw	r30,xDEBUG_SCR1(r1)
	stw	r29,xDEBUG_SCR2(r1)
	lwz	r30,xDEBUG_TRACE(r1)
	rlwinm	r30,r30,0,24,31			// 256 entries
	rlwinm	r30,r30,2,22,29
	addi	r30,r30,xDBG_TRACE_SPACE
	lis	r29,HA(_val)
	addi	r29,r29,LO(_val)
	stwx	r29,r30,r1
	lwz	r30,xDEBUG_TRACE(r1)
	addi	r30,r30,1
	rlwinm	r30,r30,0,24,31			// 256 entries
	stw	r30,xDEBUG_TRACE(r1)
	lwz	r29,xDEBUG_SCR2(r1)	
	lwz	r30,xDEBUG_SCR1(r1)
#endif
])
#define TRACE( a,b ) TRACE_VAL a,b

	
#endif   /* _H_MOLASM */
